import logging
import os
from typing import List, Union

import numpy as np
import pygame
from OpenGL.GL import glClear, GL_COLOR_BUFFER_BIT, glClearColor, glBindVertexArray, \
    glBindBuffer, GL_ARRAY_BUFFER, glCreateShader, \
    GL_VERTEX_SHADER, glShaderSource, glCompileShader, glCreateProgram, glAttachShader, glLinkProgram, glUseProgram, \
    GL_FRAGMENT_SHADER, glPixelStorei, GL_UNSIGNED_BYTE, \
    GL_RGBA, glFlush, glReadPixels, GL_PACK_ALIGNMENT, glEnable, GL_BLEND, glBlendFunc, GL_SRC_ALPHA, \
    GL_ONE_MINUS_SRC_ALPHA, glGetUniformLocation, glUniform4f
from pygame.time import Clock

from gui import GUIFacade
from gui.input import Keyboard, Mouse
from gui.layer import LayerGroup
from gui.impl.pygame import PygameKeyboard, PygameMouse
from .layer.OpenGLLayerGroup import OpenGLLayerGroup


class OpenGLGUIFacade(GUIFacade):

    def __init__(self):
        # Main properties
        self.__screenWidth = 0  # type: int
        self.__screenHeight = 0  # type: int
        self.__screenPixelWidth = 1  # type: float
        self.__screenPixelHeight = 1  # type: float
        self.__clock = None  # type: Clock
        self.__closingRequested = False  # type: bool

        # Shader properties
        self.__shaderProgramId = -1  # type: int

        # Layer groups
        self.__layerGroups = []  # type: List[Union[None,LayerGroup]]

        # Inputs
        self.__keyboard = PygameKeyboard()
        self.__mouse = PygameMouse()

    @property
    def screenWidth(self) -> int:
        return self.__screenWidth

    @property
    def screenHeight(self) -> int:
        return self.__screenHeight

    @property
    def screenPixelWidth(self) -> float:
        return self.__screenPixelWidth

    @property
    def screenPixelHeight(self) -> float:
        return self.__screenPixelHeight

    def createWindow(self, title: str, width: int, height: int):
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        pygame.init()
        pygame.display.set_mode((width, height), pygame.DOUBLEBUF | pygame.OPENGL)
        pygame.display.set_caption(title)
        glClearColor(0.0, 0.0, 0.0, 1.0)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        self.__screenWidth = width
        self.__screenHeight = height
        self.__screenPixelWidth = 2 / float(width)
        self.__screenPixelHeight = 2 / float(height)
        
    # Layer groups 

    def createLayerGroup(self, level: int = -1) -> LayerGroup:
        layerGroup = OpenGLLayerGroup(self)
        if level < 0:
            self.__layerGroups.append(layerGroup)
        elif level < len(self.__layerGroups):
            self.__layerGroups[level] = layerGroup
        else:
            self.__layerGroups.extend([None for l in range(len(self.__layerGroups), level + 1)])
            self.__layerGroups[level] = layerGroup
        return layerGroup
            
    def moveLayerGroup(self, layerGroup: Union[int, LayerGroup], level: int):
        if type(layerGroup) == LayerGroup:
            layerGroup = self.__layerGroups.index(layerGroup)
        if layerGroup < 0 or layerGroup >= len(self.__layerGroups):
            raise ValueError("Invalid level {}".format(layerGroup))
        if level < 0 or level >= len(self.__layerGroups):
            raise ValueError("Invalid level {}".format(level))
        if level != layerGroup:
            self.__layerGroups[level] = self.__layerGroups[layerGroup]
            self.__layerGroups[layerGroup] = None

    def removeLayerGroup(self, layerGroup: Union[int, LayerGroup]):
        if type(layerGroup) == LayerGroup:
            layerGroup = self.__layerGroups.index(layerGroup)
        if layerGroup < 0 or layerGroup >= len(self.__layerGroups):
            raise ValueError("Invalid level {}".format(layerGroup))
        self.__layerGroups[layerGroup] = None
        
    # Shaders

    def __createShaders(self):
        # Release focus on VBO and VAO
        glBindBuffer(GL_ARRAY_BUFFER, 0)
        glBindVertexArray(0)

        # Create vertex shader
        vertexShaderCode = '''#version 330
        layout (location=0) in vec4 vertex;
        layout (location=1) in vec2 inputUV;
                           out vec2 outputUV;
        uniform vec4 translation;
        void main() {
            gl_Position = vertex + translation;
            outputUV = inputUV;
        }'''
        vertexShaderId = glCreateShader(GL_VERTEX_SHADER)
        glShaderSource(vertexShaderId, vertexShaderCode)
        glCompileShader(vertexShaderId)

        # Create fragment shader
        fragmentShaderCode = '''#version 330
        in vec2 outputUV;
        out vec4 color;
        uniform sampler2D textureColors;
        void main() {
            color = texture(textureColors, outputUV);
        }'''
        fragmentShaderId = glCreateShader(GL_FRAGMENT_SHADER)
        glShaderSource(fragmentShaderId, fragmentShaderCode)
        glCompileShader(fragmentShaderId)

        # Create shader program
        self.__shaderProgramId = glCreateProgram()
        glAttachShader(self.__shaderProgramId, vertexShaderId)
        glAttachShader(self.__shaderProgramId, fragmentShaderId)
        glLinkProgram(self.__shaderProgramId)

    def init(self):
        self.__createShaders()
        self.__clock = pygame.time.Clock()

    def updateInputs(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.__closingRequested = True
            else:
                self.__keyboard.parsePygameEvent(event)

    def render(self):
        glClear(GL_COLOR_BUFFER_BIT)

        glUseProgram(self.__shaderProgramId)

        translationShaderVar = glGetUniformLocation(self.__shaderProgramId, "translation")
        for layerGroup in self.__layerGroups:
            if layerGroup is None:
                continue
            glUniform4f(translationShaderVar,
                        layerGroup.translationX,
                        layerGroup.translationY,
                        0.0, 0.0)
            layerGroup.draw()

        glUseProgram(0)

        pygame.display.flip()
        self.__clock.tick(60)

    @property
    def keyboard(self) -> Keyboard:
        return self.__keyboard

    @property
    def mouse(self) -> Mouse:
        return self.__mouse

    @property
    def closingRequested(self) -> bool:
        return self.__closingRequested

    @closingRequested.setter
    def closingRequested(self, b: bool):
        self.__closingRequested = b

    def captureScreenContent(self, x1: int, y1: int, x2: int, y2: int) -> np.ndarray:
        glFlush()
        glPixelStorei(GL_PACK_ALIGNMENT, 4)
        data = glReadPixels(x1, y1, x2, y2, GL_RGBA, GL_UNSIGNED_BYTE)
        capture = np.frombuffer(data, dtype=np.uint8)
        capture = capture.reshape(y2 - y1, x2 - x1, 4)
        capture = np.flip(capture, axis=0)
        return capture

    def quit(self):
        for layerGroup in self.__layerGroups:
            if layerGroup is None:
                continue
            layerGroup.deleteObjects()
        pygame.quit()
        quit()
