from typing import Union, Tuple

import numpy as np

from gui.impl.opengl import OpenGLGUIFacade
from gui.impl.opengl.private import Mesh


class OpenGLLayer:

    def __init__(self, gui: OpenGLGUIFacade):
        # Main mesh
        self.__mesh = Mesh(gui)
        self.__dynamic = True  # type: bool

        # Mesh cpu data
        self.__vertices = None  # type: Union[None, np.ndarray]
        self.__faces = None  # type: Union[None, np.ndarray]
        self.__uvMap = None  # type: Union[None, np.ndarray]

        # Transparent tile
        self.__transparentTile = (0, 0, 0, 0)  # type: Tuple[int, int, int, int]

    def setTileset(self, fileName: Union[str, np.ndarray]):
        self.__mesh.setTileset(fileName)

    def setTileSize(self, tileWidth: int, tileHeight: int):
        self.__mesh.setTileSize(tileWidth, tileHeight)
        self.__transparentTile = (
            self.__mesh.textureWidth - self.__mesh.tileWidth,
            self.__mesh.textureHeight - self.__mesh.tileHeight,
            self.__mesh.textureWidth,
            self.__mesh.textureHeight
        )

    def deleteObjects(self):
        self.__mesh.deleteObjects()

    def setRenderedFaceCount(self, count: int):
        self.__mesh.setRenderedFaceCount(count)

    @property
    def tileWidth(self) -> int:
        return self.__mesh.tileWidth

    @property
    def tileHeight(self) -> int:
        return self.__mesh.tileHeight

    @property
    def faceCount(self) -> int:
        if self.__vertices is None:
            return 0
        return self.__vertices.shape[0]

    def createFaces(self, faceCount: int) -> int:
        self.__vertices = None
        return self.addFaces(faceCount)

    def addFaces(self, faceCount: int) -> int:
        previousFaceCount = self.faceCount
        newFaceCount = previousFaceCount + faceCount

        if self.__vertices is None:
            self.__vertices = np.zeros([faceCount, 4, 2], dtype=np.float32)
            self.__uvMap = np.zeros([faceCount, 4, 2], dtype=np.float32)
            self.__faces = np.zeros([faceCount, 4], dtype=np.uint)
        else:
            self.__vertices.resize([newFaceCount, 4, 2])
            self.__uvMap.resize([newFaceCount, 4, 2])
            self.__faces.resize([newFaceCount, 4])

        for faceIndex in range(previousFaceCount, newFaceCount):
            self.setFacePixelLocation(faceIndex, 0, 0, self.__mesh.tileWidth, self.__mesh.tileHeight)
            self.setFacePixelTexture(
                faceIndex,
                self.__transparentTile[0], self.__transparentTile[1],
                self.__transparentTile[2], self.__transparentTile[3]
            )

            self.__faces[faceIndex, 0] = faceIndex * 4
            self.__faces[faceIndex, 1] = faceIndex * 4 + 1
            self.__faces[faceIndex, 2] = faceIndex * 4 + 2
            self.__faces[faceIndex, 3] = faceIndex * 4 + 3

        return previousFaceCount

    def removeFaces(self, firstIndex: int, lastIndex: int):
        assert self.__vertices is not None
        assert 0 <= firstIndex < self.__vertices.shape[0]
        assert 0 < lastIndex <= self.__vertices.shape[0]
        assert firstIndex < lastIndex

        # Delete all data related to the faces to remove
        self.__vertices = np.delete(self.__vertices, range(firstIndex, lastIndex), axis=0)
        self.__faces = np.delete(self.__faces, range(firstIndex, lastIndex), axis=0)
        self.__uvMap = np.delete(self.__uvMap, range(firstIndex, lastIndex), axis=0)

        # Update the face indexes
        faceCount = self.__faces.shape[0]
        for faceIndex in range(firstIndex, faceCount):
            self.__faces[faceIndex, 0] = faceIndex * 4
            self.__faces[faceIndex, 1] = faceIndex * 4 + 1
            self.__faces[faceIndex, 2] = faceIndex * 4 + 2
            self.__faces[faceIndex, 3] = faceIndex * 4 + 3

    def setFacePixelLocation(self, faceIndex: int, x1: float, y1: float, x2: float, y2: float):
        assert 0 <= faceIndex < self.faceCount

        spriteScreenX1 = -1 + x1 * self.__mesh.screenPixelWidth
        spriteScreenY1 = 1 - y1 * self.__mesh.screenPixelHeight
        spriteScreenX2 = -1 + x2 * self.__mesh.screenPixelWidth
        spriteScreenY2 = 1 - y2 * self.__mesh.screenPixelHeight
        self.__vertices[faceIndex, 0] = [spriteScreenX1, spriteScreenY2]
        self.__vertices[faceIndex, 1] = [spriteScreenX1, spriteScreenY1]
        self.__vertices[faceIndex, 2] = [spriteScreenX2, spriteScreenY1]
        self.__vertices[faceIndex, 3] = [spriteScreenX2, spriteScreenY2]

    def setFacePixelTexture(self, faceIndex: int, tileX1: float, tileY1: float, tileX2: float, tileY2: float):
        assert 0 <= faceIndex < self.faceCount

        spriteTextureX1 = tileX1 / self.__mesh.textureWidth
        spriteTextureY1 = tileY1 / self.__mesh.textureHeight
        spriteTextureX2 = tileX2 / self.__mesh.textureWidth
        spriteTextureY2 = tileY2 / self.__mesh.textureHeight
        self.__uvMap[faceIndex, 0] = [spriteTextureX1, spriteTextureY2]
        self.__uvMap[faceIndex, 1] = [spriteTextureX1, spriteTextureY1]
        self.__uvMap[faceIndex, 2] = [spriteTextureX2, spriteTextureY1]
        self.__uvMap[faceIndex, 3] = [spriteTextureX2, spriteTextureY2]

    def setFaceTileLocation(self, faceIndex: int, x: float, y: float):
        assert 0 <= faceIndex < self.__vertices.shape[0]

        spriteScreenX1 = -1 + x * self.__mesh.screenTileWidth
        spriteScreenY1 = 1 - y * self.__mesh.screenTileHeight
        spriteScreenX2 = spriteScreenX1 + self.__mesh.screenTileWidth
        spriteScreenY2 = spriteScreenY1 - self.__mesh.screenTileHeight
        self.__vertices[faceIndex, 0] = [spriteScreenX1, spriteScreenY2]
        self.__vertices[faceIndex, 1] = [spriteScreenX1, spriteScreenY1]
        self.__vertices[faceIndex, 2] = [spriteScreenX2, spriteScreenY1]
        self.__vertices[faceIndex, 3] = [spriteScreenX2, spriteScreenY2]

    def setFaceTileTexture(self, faceIndex: int, tileX: float, tileY: float):
        assert 0 <= faceIndex < self.__vertices.shape[0]

        spriteTextureX1 = tileX * self.__mesh.textureTileWidth
        spriteTextureY1 = tileY * self.__mesh.textureTileHeight
        spriteTextureX2 = spriteTextureX1 + self.__mesh.textureTileWidth
        spriteTextureY2 = spriteTextureY1 + self.__mesh.textureTileHeight
        self.__uvMap[faceIndex, 0] = [spriteTextureX1, spriteTextureY2]
        self.__uvMap[faceIndex, 1] = [spriteTextureX1, spriteTextureY1]
        self.__uvMap[faceIndex, 2] = [spriteTextureX2, spriteTextureY1]
        self.__uvMap[faceIndex, 3] = [spriteTextureX2, spriteTextureY2]

    def hideFace(self, faceIndex: int):
        self.setFacePixelTexture(
            faceIndex,
            self.__transparentTile[0], self.__transparentTile[1],
            self.__transparentTile[2], self.__transparentTile[3]
        )

    def sendData(self, dynamic: bool = False):
        self.__dynamic = dynamic
        self.__mesh.setData(self.__vertices, self.__faces, self.__uvMap, dynamic)

    def draw(self):
        if self.__dynamic:
            self.__mesh.updateData(self.__vertices, self.__faces, self.__uvMap)
        self.__mesh.draw()
