import logging
from typing import Union, Dict

import numpy as np
import pygame

from gui.text import TextStyle
from gui.impl.pygame import surface2numpy
from gui.impl.pygame.text import CharacterTile, CharacterRendererFactory, CharacterRenderer


class TextTileset:
    def __init__(self):
        self.__characterRendererFactory = CharacterRendererFactory.getInstance()

        self.__tileset = None  # type: Union[None, pygame.Surface]
        self.__nextTileX = 0  # type: int
        self.__nextTileY = 0  # type: int
        self.__currentRowHeight = 0  # type: int
        self.__characterTiles = {}  # type: Dict[TextStyle,Dict[str, CharacterTile]]

        self.__currentCharacterRenderer = None  # type: Union[None, CharacterRenderer]
        self.__currentStyle = None  # type: Union[None, TextStyle]
        self.__currentCharacterTiles = {}  # type: Dict[str, CharacterTile]

    def __resize(self, width: int, height: int):
        # GPUs prefer square image size with a power of two
        def nextPowerOfTwo(x: int) -> int:
            return 1 if x == 0 else 2 ** (x - 1).bit_length()

        imageWidth = nextPowerOfTwo(width)
        imageHeight = nextPowerOfTwo(height)
        imageSize = max(imageWidth, imageHeight)
        self.__tileset = pygame.Surface((imageSize, imageSize), flags=pygame.SRCALPHA)

        # Copy characters in the new tileset
        self.__nextTileX = 0
        self.__nextTileY = 0
        self.__currentRowHeight = 0
        for style in self.__characterTiles:
            self.__currentCharacterTiles = self.__characterTiles[style]
            currentCharacters = "".join(self.__currentCharacterTiles.keys())
            self.__currentCharacterTiles.clear()
            self.__currentCharacterRenderer = self.__characterRendererFactory.create(style)
            self.addCharacters(currentCharacters)
        self.setStyle(self.__currentStyle)

    def setStyle(self, style: TextStyle):
        self.__currentCharacterRenderer = self.__characterRendererFactory.create(style)
        if style not in self.__characterTiles:
            self.__characterTiles[style] = {}
        self.__currentStyle = style
        self.__currentCharacterTiles = self.__characterTiles[style]
        if self.__tileset is None:
            self.__resize(256, 256)

    def getStyle(self) -> Union[None, TextStyle]:
        return self.__currentStyle

    def addCharacters(self, characters: str) -> bool:
        tilesetUpdated = False
        for character in characters:
            if character in self.__currentCharacterTiles:
                continue
            if character == "\n":
                continue
            tilesetUpdated = True

            # Render the character
            surface = self.__currentCharacterRenderer.render(character)
            tileDefaultHeight = self.__currentCharacterRenderer.tileDefaultHeight
            characterWidth = surface.get_width()
            if tileDefaultHeight < surface.get_height():
                logging.warning("The height of character {} is higher than the default.".format(character))
            if tileDefaultHeight > self.__currentRowHeight:
                self.__currentRowHeight = tileDefaultHeight

            # Ensure that next tile location in inside the tileset
            if (self.__nextTileX + characterWidth) > self.__tileset.get_width():
                self.__nextTileX = 0
                self.__nextTileY += self.__currentRowHeight
            if (self.__nextTileY + self.__currentRowHeight) > self.__tileset.get_height():
                self.__resize(2 * self.__tileset.get_width(), 2 * self.__tileset.get_height())
                if (self.__nextTileX + characterWidth) > self.__tileset.get_width():
                    self.__nextTileX = 0
                    self.__nextTileY += self.__currentRowHeight

            # Blit the character in the tileset
            self.__tileset.blit(surface, (self.__nextTileX, self.__nextTileY))

            # Compute and save the tile location
            self.__currentCharacterTiles[character] = CharacterTile(
                self.__nextTileX, self.__nextTileY,
                self.__nextTileX + characterWidth, self.__nextTileY + tileDefaultHeight,
                self.__currentCharacterRenderer.getCharacterAdvance(character)
            )

            # Update tile next location
            self.__nextTileX += characterWidth
        return tilesetUpdated

    def toNumpyArray(self) -> np.ndarray:
        return surface2numpy(self.__tileset)

    def getCharacterTile(self, character: str) -> CharacterTile:
        return self.__currentCharacterTiles[character]


