from typing import Union

import numpy as np

from gui import GUIFacade
from gui.input import Keyboard
from gui.layer import GridLayer
from gui.text import TextStyle
from logic import Logic
from logic.command import CreateRegionCommand, SetGroundValueCommand, FillGroundValueCommand
from state import Region, StateListener, State
from state.Region import GROUND_GRASS, GROUND_SWAMP, GROUND_DIRT, GROUND_SAND, GROUND_DEAD_LAND
from state.io import MainIO
from .GameMode import GameMode


class EditGameMode(GameMode, StateListener):

    def __init__(self, gui: GUIFacade):
        self.__gui = gui
        self.__state = State()
        self.__state.addListener(self)
        self.__logic = Logic()

        self.__tileSize = 32  # type: int
        self.__viewX = 0  # type: int
        self.__viewY = 0  # type: int
        self.__displayBorders = True

        self.__worldLayerGroup = gui.createLayerGroup()
        self.__groundLayer = None  # type: Union[None, GridLayer]
        self.__currentRegionName = None  # type: Union[None, str]
        self.__currentRegion = None  # type: Union[None, Region]
        self.__currentBrush = GROUND_SWAMP

        self.__frontLayerGroup = gui.createLayerGroup()
        self.__framesLayer = self.__frontLayerGroup.createUILayer()
        self.__framesLayer.setTileset("assets/ui/ui.png")
        self.__framesLayer.setTileSize(32, 32)
        self.__framesLayer.createFrameBox(
            self.__tileSize,
            gui.screenHeight - self.__tileSize * 8,
            gui.screenWidth - self.__tileSize * 21,
            gui.screenHeight - self.__tileSize,
        )
        style = TextStyle("assets/font/terminal-grotesque.ttf", self.__tileSize, (255, 255, 255))
        self.__textLayer = self.__frontLayerGroup.createTextLayer()
        self.__textLayer.setStyle(style)
        self.__textLayer.setMaxCharacterCount(500)
        self.__textLayer.setText(
            self.__tileSize * 1.5,
            gui.screenHeight - self.__tileSize * 7.5,
            "<u><b>Commands</b></u>\n"
            "<b>Left button:</b> Paint\n"
            "<b>Right button:</b> Fill\n"
            "<b>F1:</b> Load\n"
            "<b>F2:</b> Save\n"
        )
        self.__textLayer = self.__frontLayerGroup.createTextLayer()
        self.__textLayer.setStyle(style)
        self.__textLayer.setMaxCharacterCount(500)
        self.__textLayer.setText(
            self.__tileSize * 10.5,
            gui.screenHeight - self.__tileSize * 7.5,
            "\n"
            "<b>1:</b> Select grass\n"
            "<b>2:</b> Select swamp\n"
            "<b>3:</b> Select dirt\n"
            "<b>4:</b> Select sand\n"
            "<b>5:</b> Select dead land\n"
        )

        self.__coordsTextLayer = self.__frontLayerGroup.createTextLayer()
        self.__coordsTextLayer.setStyle(style)
        self.__coordsTextLayer.setMaxCharacterCount(500)
        self.__coordsTextLayer.setText(
            self.__tileSize * 1.5,
            gui.screenHeight - self.__tileSize * 2.5,
            "Coordinates: (x, y)"
        )

    #############################################################
    # Mode
    #############################################################

    def init(self):
        self.__logic.addCommand(CreateRegionCommand(self.__state, 'test', 64, 48))

    def handleInputs(self) -> bool:
        # Mouse
        mouse = self.__gui.mouse
        mouseX = (mouse.x + self.__viewX) // self.__tileSize
        mouseY = (mouse.y + self.__viewY) // self.__tileSize

        self.__coordsTextLayer.setText(
            self.__tileSize * 1.5,
            self.__gui.screenHeight - self.__tileSize * 2.5,
            "Coordinates: ({}, {})".format(mouseX, mouseY)
        )

        if mouse.button1:
            self.__logic.addCommand(SetGroundValueCommand(self.__state, self.__currentRegionName, mouseX, mouseY, self.__currentBrush))
            return True
        if mouse.button3:
            if 0 <= mouseX < self.__currentRegion.width and 0 <= mouseY < self.__currentRegion.height:
                value = self.__currentRegion.getGroundValue(mouseX, mouseY)
                self.__logic.addCommand(FillGroundValueCommand(self.__state, self.__currentRegionName, mouseX, mouseY, value, self.__currentBrush))
            return True

        # Keyboard
        keyboard = self.__gui.keyboard
        shiftX = 0
        shiftY = 0
        if keyboard.isKeyPressed(Keyboard.K_LEFT):
            shiftX -= self.__tileSize
        if keyboard.isKeyPressed(Keyboard.K_RIGHT):
            shiftX += self.__tileSize
        if keyboard.isKeyPressed(Keyboard.K_UP):
            shiftY -= self.__tileSize
        if keyboard.isKeyPressed(Keyboard.K_DOWN):
            shiftY += self.__tileSize
        if shiftX != 0 or shiftY != 0:
            self.__viewX += shiftX
            self.__viewY += shiftY
            self.__worldLayerGroup.setTranslation(self.__viewX, self.__viewY)
            return True

        if keyboard.isKeyPressed(Keyboard.K_1):
            self.__currentBrush = GROUND_GRASS
            return True
        if keyboard.isKeyPressed(Keyboard.K_2):
            self.__currentBrush = GROUND_SWAMP
            return True
        if keyboard.isKeyPressed(Keyboard.K_3):
            self.__currentBrush = GROUND_DIRT
            return True
        if keyboard.isKeyPressed(Keyboard.K_4):
            self.__currentBrush = GROUND_SAND
            return True
        if keyboard.isKeyPressed(Keyboard.K_5):
            self.__currentBrush = GROUND_DEAD_LAND
            return True

        if keyboard.isKeyPressed(Keyboard.K_F1):
            self.__state = MainIO.load("level.bin")
            self.__state.addListener(self)
            self.currentRegionChanged(self.__state, self.__currentRegionName)
            return True
        if keyboard.isKeyPressed(Keyboard.K_F2):
            MainIO.save("level.bin", self.__state)
            return True

        return False

    def updateState(self):
        self.__logic.executeCommands()

    #############################################################
    # Events handling
    #############################################################

    def currentRegionChanged(self, state: State, regionName: str):
        # Case where we remove the region
        if regionName is None:
            self.__worldLayerGroup.removeLayer(self.__groundLayer)
            self.__groundLayer = None
            self.__currentRegionName = None
            self.__currentRegion = None
            return

        self.__currentRegionName = regionName
        region = state.getRegion(regionName)
        self.__currentRegion = region
        width = region.width
        height = region.height

        # Create the ground layer
        self.__groundLayer = self.__worldLayerGroup.createGridLayer(0)
        self.__groundLayer.setTileset("assets/level/groundcombi.png")
        self.__groundLayer.setTileSize(self.__tileSize, self.__tileSize)
        self.__groundLayer.setTiles(np.zeros([width, height, 2], dtype=np.int32))
        for y in range(height):
            for x in range(width):
                self.__updateRegionTile(region, x, y)

    def regionCellChanged(self, state: State, regionName: str, x: int, y: int):
        if regionName != self.__currentRegionName:
            return
        assert self.__groundLayer is not None

        region = self.__currentRegion
        for j in range(3):
            for i in range(3):
                xp = x + i - 1
                yp = y + j - 1
                self.__updateRegionTile(region, xp, yp)

    def __updateRegionTile(self, region: Region, x: int, y: int):
        if x < 0 or x >= region.width or y < 0 or y >= region.height:
            return

        neightbors = region.getGroundNeightbors(x, y)
        value = neightbors[1, 1]
        otherValue = min(value, neightbors[1, 0], neightbors[1, 2], neightbors[0, 1], neightbors[2, 1])
        if otherValue == value:
            otherValue = neightbors.min()

        tilesetLocation = self.groundTilesetLocations[value, otherValue]
        tileX = tilesetLocation[0] * 8
        tileY = tilesetLocation[1] * 6
        if value == otherValue or not self.__displayBorders:
            tileX += 6
            tileY += 1
        else:
            mask = neightbors != otherValue
            if mask[0, 0] and (not mask[0, 1] or not mask[1, 0]):
                mask[0, 0] = 0
            if mask[0, 2] and (not mask[0, 1] or not mask[1, 2]):
                mask[0, 2] = 0
            if mask[2, 0] and (not mask[1, 0] or not mask[2, 1]):
                mask[2, 0] = 0
            if mask[2, 2] and (not mask[1, 2] or not mask[2, 1]):
                mask[2, 2] = 0
            powers = np.array([1, 2, 4, 8, 0, 16, 32, 64, 128], dtype=np.int32)
            code = (mask.flatten() * powers).sum()

            shift = self.code2tile[code]
            tileX += shift[0]
            tileY += shift[1]

        self.__groundLayer.setTile(x, y, tileX, tileY)

    groundTilesetLocations = np.array([
        [[7, 7], [7, 7], [7, 7], [7, 7], [7, 7], [7, 7]],  # Invalid
        [[7, 7], [0, 0], [0, 0], [1, 0], [2, 0], [3, 0]],  # Grass
        [[7, 7], [4, 0], [4, 0], [5, 0], [6, 0], [7, 0]],  # Swamp
        [[7, 7], [0, 1], [1, 1], [1, 1], [2, 1], [3, 1]],  # Dirt
        [[7, 7], [4, 1], [5, 1], [6, 1], [6, 1], [7, 1]],  # Sand
        [[7, 7], [0, 2], [1, 2], [2, 2], [3, 2], [3, 2]],  # Dead land
    ], dtype=np.int32)

    code2tile = {
        # Code 0
        # 000
        # 000
        # 000
        0: [0, 0],

        # Code 2
        # 000
        # 100
        # 000
        2: [3, 0],

        # Code 8
        # 010
        # 000
        # 000
        8: [4, 2],

        # Code 10
        # 010
        # 100
        # 000
        10: [1, 2],

        # Code 11
        # 110
        # 100
        # 000
        11: [7, 2],

        # Code 16
        # 000
        # 000
        # 010
        16: [4, 0],

        # Code 18
        # 000
        # 100
        # 010
        18: [1, 1],

        # Code 22
        # 000
        # 100
        # 110
        22: [7, 0],

        # Code 24
        # 010
        # 000
        # 010
        24: [4, 1],

        # Code 26
        # 010
        # 100
        # 010
        26: [3, 2],

        # Code 27
        # 110
        # 100
        # 010
        27: [1, 3],

        # Code 30
        # 010
        # 100
        # 110
        30: [1, 4],

        # Code 31
        # 110
        # 100
        # 110
        31: [7, 1],

        # Code 64
        # 000
        # 001
        # 000
        64: [1, 0],

        # Code 66
        # 000
        # 101
        # 000
        66: [2, 0],

        # Code 72
        # 010
        # 001
        # 000
        72: [0, 2],

        # Code 74
        # 010
        # 101
        # 000
        74: [2, 2],

        # Code 75
        # 110
        # 101
        # 000
        75: [2, 4],

        # Code 80
        # 000
        # 001
        # 010
        80: [0, 1],

        # Code 82
        # 000
        # 101
        # 010
        82: [3, 1],

        # Code 86
        # 000
        # 101
        # 110
        86: [2, 3],

        # Code 88
        # 010
        # 001
        # 010
        88: [2, 1],

        # Code 90
        # 010
        # 101
        # 010
        90: [6, 5],

        # Code 91
        # 110
        # 101
        # 010
        91: [6, 3],

        # Code 94
        # 010
        # 101
        # 110
        94: [6, 4],

        # Code 95
        # 110
        # 101
        # 110
        95: [2, 5],

        # Code 104
        # 011
        # 001
        # 000
        104: [5, 2],

        # Code 106
        # 011
        # 101
        # 000
        106: [3, 4],

        # Code 107
        # 111
        # 101
        # 000
        107: [6, 2],

        # Code 120
        # 011
        # 001
        # 010
        120: [0, 3],

        # Code 122
        # 011
        # 101
        # 010
        122: [7, 3],

        # Code 123
        # 111
        # 101
        # 010
        123: [1, 5],

        # Code 126
        # 011
        # 101
        # 110
        126: [5, 5],

        # Code 127
        # 111
        # 101
        # 110
        127: [4, 3],

        # Code 208
        # 000
        # 001
        # 011
        208: [5, 0],

        # Code 210
        # 000
        # 101
        # 011
        210: [3, 3],

        # Code 214
        # 000
        # 101
        # 111
        214: [6, 0],

        # Code 216
        # 010
        # 001
        # 011
        216: [0, 4],

        # Code 218
        # 010
        # 101
        # 011
        218: [7, 4],

        # Code 219
        # 110
        # 101
        # 011
        219: [4, 5],

        # Code 222
        # 010
        # 101
        # 111
        222: [0, 5],

        # Code 223
        # 110
        # 101
        # 111
        223: [4, 4],

        # Code 248
        # 011
        # 001
        # 011
        248: [5, 1],

        # Code 250
        # 011
        # 101
        # 011
        250: [3, 5],

        # Code 251
        # 111
        # 101
        # 011
        251: [5, 3],

        # Code 254
        # 011
        # 101
        # 111
        254: [5, 4],

        # Code 255
        # 111
        # 101
        # 111
        255: [6, 1],
    }